
from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from ..common import *
from .. import config as config
import json
import math


GRID_SIZE = 20


def round(value, base):
	return math.floor(value / base + 0.5) * base
	
def alignToGrid(pos):
	x = round(pos.x(), GRID_SIZE)
	y = round(pos.y(), GRID_SIZE)
	return QPointF(x, y)
	
def mergeColors(c1, c2, p=.5):
	r = c1.red() * (1 - p) + c2.red() * p
	g = c1.green() * (1 - p) + c2.green() * p
	b = c1.blue() * (1 - p) + c2.blue() * p
	return QColor(r, g, b)
	

ShapeColors = {
	"black": QColor(0, 0, 0),
	"white": QColor(255, 255, 255),
	
	"red": QColor(255, 0, 0),
	"green": QColor(0, 255, 0),
	"blue": QColor(0, 0, 255),
	
	"lightred": QColor(255, 128, 128),
	"lightblue": QColor(128, 128, 255),
	
	"yellow": QColor(255, 255, 0),
	
	"orange": QColor(255, 128, 0),
	
	"lightorange": QColor(255, 192, 128)
}

class ShapeElement:
	def __init__(self, type, **data):
		self.type = type
		if self.type == "line":
			self.x1, self.y1 = data["x1"], data["y1"]
			self.x2, self.y2 = data["x2"], data["y2"]
			self.curve = data["curve"]
		elif self.type == "arc":
			self.x, self.y = data["x"], data["y"]
			self.w, self.h = data["w"], data["h"]
			self.start, self.span = data["start"], data["span"]
		elif self.type == "circle":
			self.x, self.y = data["x"], data["y"]
			self.r = data["r"]
		elif self.type == "rect":
			self.x, self.y = data["x"], data["y"]
			self.w, self.h = data["w"], data["h"]
		elif self.type == "arrow":
			self.x1, self.y1 = data["x1"], data["y1"]
			self.x2, self.y2 = data["x2"], data["y2"]
			self.stretch = data["stretch"]
			self.curve = data["curve"]
		elif self.type == "polygon":
			self.points = data["points"]
		else:
			raise ValueError("Unknown shape element type: %s" %self.type)
			
			
class ShapePart:
	def __init__(self):
		self.pen = QPen()
		self.brush = QBrush()
		self.path = QPainterPath()
		self.elements = []
		self.stroke = 0
		
	def setPen(self, pen): self.pen = pen
	def setBrush(self, brush): self.brush = brush
	def setStroke(self, stroke): self.stroke = stroke
	
	def addElement(self, element):
		self.elements.append(element)
		
	def load(self, data):
		pen = data.get("pen")
		if pen:
			self.pen = QPen(ShapeColors[pen["color"]])
			self.pen.setWidth(pen["width"])
			self.pen.setCapStyle(Qt.RoundCap)
		
		brush = data.get("brush")
		if brush:
			self.brush = QBrush(ShapeColors[brush["color"]])
		
		self.elements = []
		for element in data["elements"]:
			self.elements.append(ShapeElement(**element))
		
	def update(self):
		self.path = QPainterPath()
		
		for element in self.elements:
		
			if element.type == "line":
				dx = element.x2 - element.x1
				dy = element.y2 - element.y1
				angle = math.atan2(dy, dx)
				
				centerx = (element.x1 + element.x2) / 2
				centery = (element.y1 + element.y2) / 2
				
				controlx = centerx + math.cos(angle + math.pi / 2) * element.curve
				controly = centery + math.sin(angle + math.pi / 2) * element.curve
				
				self.path.moveTo(element.x1, element.y1)
				self.path.quadTo(controlx, controly, element.x2, element.y2)
				self.path.quadTo(controlx, controly, element.x1, element.y1)
			
			elif element.type == "arc":
				self.path.arcMoveTo(element.x, element.y, element.w, element.h, element.start)
				self.path.arcTo(element.x, element.y, element.w, element.h, element.start, element.span)
			
			elif element.type == "circle":
				self.path.addEllipse(QPointF(element.x, element.y), element.r, element.r)
			
			elif element.type == "rect":
				self.path.addRect(element.x, element.y, element.w, element.h)
				
			elif element.type == "polygon":
				points = []
				for point in element.points:
					points.append(QPointF(point[0], point[1]))
				self.path.addPolygon(QPolygonF(points))
			
			elif element.type == "arrow":
				dx = element.x2 - element.x1
				dy = element.y2 - element.y1
				length = math.sqrt(dx * dx + dy * dy)
				
				angle1 = math.atan2(dy, dx)
				angle2 = angle1 - math.atan2(element.curve, length / 2)
				
				centerx = (element.x1 + element.x2) / 2
				centery = (element.y1 + element.y2) / 2
				
				controlx = centerx + math.cos(angle1 + math.pi / 2) * element.curve
				controly = centery + math.sin(angle1 + math.pi / 2) * element.curve
				
				self.path.moveTo(element.x1, element.y1)
				self.path.quadTo(controlx, controly, element.x2, element.y2)
				self.path.quadTo(controlx, controly, element.x1, element.y1)
				
				self.path.moveTo(element.x2, element.y2)
				self.path.lineTo(
					element.x2 + element.stretch * math.cos(angle2 + math.pi * .75),
					element.y2 + element.stretch * math.sin(angle2 + math.pi * .75)
				)
				
				self.path.moveTo(element.x2, element.y2)
				self.path.lineTo(
					element.x2 + element.stretch * math.cos(angle2 - math.pi * .75),
					element.y2 + element.stretch * math.sin(angle2 - math.pi * .75)
				)
		
		if self.stroke > 0:
			stroker = QPainterPathStroker()
			stroker.setWidth(self.stroke)
			shapePath = stroker.createStroke(self.path)
			self.shapePath = self.path.united(shapePath)
		else:
			self.shapePath = self.path
	
	
class Shape:
	def __init__(self):
		self.parts = []
		self.path = QPainterPath()
		self.shapePath = QPainterPath()
		self.rect = QRectF()
		
	def addPart(self, part):
		self.parts.append(part)
		self.update()
		
	def load(self, data):
		self.parts = []
		for part in data:
			shapePart = ShapePart()
			shapePart.load(part)
			self.parts.append(shapePart)
			
		self.update()
			
	def draw(self, painter, filter=None):
		painter.setRenderHint(QPainter.Antialiasing)
		
		for part in self.parts:
			painter.save()
			
			pen = QPen(part.pen)
			if filter:
				filter.applyToPen(pen)
			painter.setPen(pen)
				
			brush = QBrush(part.brush)
			if filter:
				filter.applyToBrush(brush)
			painter.setBrush(brush)
			
			painter.drawPath(part.path)
			painter.restore()
		
	def update(self):
		self.path = QPainterPath()
		self.shapePath = QPainterPath()
		for part in self.parts:
			part.update()
			self.path.addPath(part.path)
			self.shapePath.addPath(part.shapePath)
		self.rect = self.shapePath.boundingRect()
		
		
class Style:
	def __init__(self):
		self.shapes = {}
		
	def load(self, filename):
		with open(filename) as f:
			info = json.load(f)
		
		for name, data in info["shapes"].items():
			shape = Shape()
			shape.load(data)
			self.shapes[name] = shape

			
class DragMode:
	NONE = 0
	NORMAL = 1
	SPECIAL = 2

	
class EditorItem(QGraphicsItem):
	def __init__(self, scene):
		super().__init__()
		self.setFlag(QGraphicsItem.ItemIsSelectable)
		
		self.signals = SignalListener()
		
		self.scene = scene
		self.dragMode = DragMode.NONE
		self.invalid = False
		
		self.doubleClicked = Signal()
	
	def connect(self, signal, callback):
		self.signals.connect(signal, callback)
	
	def disconnect(self):
		self.signals.disconnect()
	
	def setInvalid(self, invalid):
		if self.invalid != invalid:
			self.invalid = invalid
			self.update()
			
	def drag(self, param): pass
	def delete(self): pass
	
	def removeFromScene(self):
		self.scene.removeItem(self)
		
	def addToScene(self):
		self.scene.addItem(self)
		
	def checkCollisions(self): pass
	
	
class ShapeFilter:
	def __init__(self, item):
		self.item = item

	def applyToPen(self, pen):
		if self.item.isSelected():
			pen.setColor(Qt.blue)
		if self.item.hover:
			color = mergeColors(pen.color(), QColor(Qt.gray))
			pen.setColor(color)

	def applyToBrush(self, brush):
		if self.item.hover:
			color = mergeColors(brush.color(), QColor(Qt.gray))
			brush.setColor(color)
	
	
class EditorShape(EditorItem):
	def __init__(self, scene, shape=None):
		super().__init__(scene)
		self.hover = False
		self.filter = ShapeFilter(self)
		
		self.shp = shape
		if self.shp is None:
			self.shp = Shape()
	
	def setShape(self, shape):
		self.shp = shape
		self.updateShape()
		
	def updateShape(self):
		self.prepareGeometryChange()
		self.shp.update()
		self.update()
		
	def setHover(self, hover):
		if self.hover != hover:
			self.hover = hover
			self.update()
		
	def checkHover(self, pos):
		pos = self.mapFromScene(pos)
		self.setHover(self.contains(pos))
		
	def shape(self):
		return self.shp.shapePath
		
	def boundingRect(self):
		return self.shp.rect.adjusted(-2, -2, 2, 2)
		
	def paint(self, painter, option, widget):
		self.shp.draw(painter, self.filter)
	
	
class DragParam:
	def __init__(self, pos, mouse):
		self.pos = pos
		self.mouse = mouse

	
class ObjectDragger:
	def __init__(self):
		self.reset()
	
	def reset(self):
		self.items = []
		self.itemBase = []
		self.dragBase = None
		
	def isDragging(self):
		return self.items and self.dragBase
		
	def init(self, pos, items):
		self.dragBase = pos
		
		for item in items:
			item.setZValue(item.zValue() + .5)
		self.items = items
		self.itemBase = [item.pos() for item in items]
		
	def update(self, pos):
		if self.isDragging():
			posDiff = pos - self.dragBase
			for item, base in zip(self.items, self.itemBase):
				item.drag(DragParam(base + posDiff, pos))
			for item in self.items:
				item.checkCollisions()
	
	def finish(self, pos):
		if any(item.invalid for item in self.items):
			for item, base in zip(self.items, self.itemBase):
				item.drag(DragParam(base, self.dragBase))
				item.setInvalid(False)
		
		for item in self.items:
			item.setZValue(item.zValue() - .5)
		
		self.reset()
		
	def removeItem(self, item):
		if item in self.items:
			index = self.items.index(item)
			self.items.pop(index)
			self.itemBase.pop(index)


class EditorScene(QGraphicsScene):
	def __init__(self):
		super().__init__(-10000, -10000, 20000, 20000)
		
		self.dragger = ObjectDragger()
		self.placedItem = None
		
		self.hoverEnabled = True
		self.gridEnabled = True
		
		self.controller = None
		
	def cleanup(self):
		for item in self.items():
			item.disconnect()
		
	def setController(self, controller):
		self.controller = controller
		
	def selectAll(self):
		for item in self.items():
			item.setSelected(True)
			
	def findItem(self, pos, *classes):
		for item in self.items(pos):
			if isinstance(item, classes):
				return item
				
	def updateHover(self, pos):
		enabled = self.hoverEnabled and not self.dragger.isDragging()
		for item in self.items():
			if isinstance(item, EditorShape):
				if enabled:
					item.checkHover(pos)
				else:
					item.setHover(False)
			
	def setHoverEnabled(self, hover):
		self.hoverEnabled = hover
		
	def setGridEnabled(self, grid):
		self.gridEnabled = grid
		self.update()
			
	def drawBackground(self, painter, rect):
		pen = QPen()
		pen.setColor(QColor(230, 230, 230))
		painter.setPen(pen)
		
		if self.gridEnabled:
			for x in range(int(rect.left()) // GRID_SIZE, int(rect.right()) // GRID_SIZE + 1):
				painter.drawLine(x * GRID_SIZE, rect.top(), x * GRID_SIZE, rect.bottom())
			for y in range(int(rect.top()) // GRID_SIZE, int(rect.bottom()) // GRID_SIZE + 1):
				painter.drawLine(rect.left(), y * GRID_SIZE, rect.right(), y * GRID_SIZE)
			
	def keyPressEvent(self, e):
		super().keyPressEvent(e)
		if e.key() == Qt.Key_Delete:
			for item in self.selectedItems():
				self.dragger.removeItem(item)
				item.delete()
			if self.placedItem:
				self.placedItem.checkCollisions()
			
	def mousePressEvent(self, e):
		pos = e.scenePos()
		if self.dragger.isDragging() or self.placedItem:
			e.accept()
		else:
			super().mousePressEvent(e)
			if e.button() == Qt.LeftButton:
				item = self.findItem(pos, EditorItem)
				if item:
					if item.dragMode == DragMode.NORMAL:
						items = [i for i in self.selectedItems() if i.dragMode == DragMode.NORMAL]
						self.dragger.init(pos, items)
					elif item.dragMode == DragMode.SPECIAL:
						self.dragger.init(pos, [item])
			elif e.button() == Qt.RightButton:
				self.placedItem = self.controller.startPlacement(pos)
				if self.placedItem:
					self.placedItem.checkCollisions()
					self.addItem(self.placedItem)
				e.accept()
					
	def mouseMoveEvent(self, e):
		super().mouseMoveEvent(e)
		
		self.updateHover(e.scenePos())
		
		self.dragger.update(e.scenePos())
		if self.placedItem:
			self.placedItem.drag(DragParam(e.scenePos(), e.scenePos()))
			self.placedItem.checkCollisions()
			
	def mouseReleaseEvent(self, e):
		super().mouseReleaseEvent(e)
		if e.button() == Qt.LeftButton:
			self.dragger.finish(e.scenePos())
		elif e.button() == Qt.RightButton:
			if self.placedItem:
				self.removeItem(self.placedItem)
				self.controller.finishPlacement(e.scenePos(), self.placedItem)
				self.placedItem = None

	def mouseDoubleClickEvent(self, e):
		pos = e.scenePos()
		if e.button() == Qt.LeftButton:
			item = self.findItem(pos, EditorItem)
			if item:
				item.doubleClicked.emit()
		

class EditorView(QGraphicsView):
	def __init__(self, scene):
		super().__init__(scene)
		self.setMouseTracking(True)
		self.setDragMode(QGraphicsView.RubberBandDrag)
		self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
		self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
		self.horizontalScrollBar().disconnect()
		self.verticalScrollBar().disconnect()
		
		self.prevPos = None
		self.prevDelta = None
		self.handDrag = False
		
		self.zoom = 1
		
		self.scrollStep = config.get("ui.keyboard_scroll_speed")
		
	def updateDragMode(self):
		if self.handDrag:
			self.setDragMode(QGraphicsView.ScrollHandDrag)
		else:
			self.setDragMode(QGraphicsView.RubberBandDrag)
		
	def setHandDrag(self, enabled):
		self.handDrag = enabled
		if self.prevPos is None:
			self.updateDragMode()
		
	def mousePressEvent(self, e):
		super().mousePressEvent(e)
		if e.button() == Qt.LeftButton:
			if not self.itemAt(e.pos()):
				self.prevPos = self.mapToScene(e.pos())
				self.prevDelta = QPointF()
		
	def mouseMoveEvent(self, e):
		super().mouseMoveEvent(e)
		if self.dragMode() == QGraphicsView.ScrollHandDrag:
			if self.prevPos is not None:
				pos = self.mapToScene(e.pos())
				delta = pos - self.prevPos + self.prevDelta
				self.translate(delta.x(), delta.y())
				self.prevPos = pos
				self.prevDelta = delta
		
	def mouseReleaseEvent(self, e):
		super().mouseReleaseEvent(e)
		if e.button() == Qt.LeftButton:
			self.updateDragMode()
			self.prevPos = None
			self.prevDelta = None
		
	def keyPressEvent(self, e):
		super().keyPressEvent(e)
		
		key = e.key()
		if key == Qt.Key_Left:
			self.translate(self.scrollStep, 0)
		elif key == Qt.Key_Right:
			self.translate(-self.scrollStep, 0)
		elif key == Qt.Key_Up:
			self.translate(0, self.scrollStep)
		elif key == Qt.Key_Down:
			self.translate(0, -self.scrollStep)
		
	def wheelEvent(self, e):
		zoom = 1.0008 ** e.angleDelta().y()
		newZoom = self.zoom * zoom
		if newZoom > 0.1 and newZoom < 10:
			self.zoom = newZoom
			prevPos = self.mapToScene(e.pos())
			self.scale(zoom, zoom)
			newPos = self.mapToScene(e.pos())
			delta = newPos - prevPos
			self.translate(delta.x(), delta.y())



from PyQt5.QtCore import QXmlStreamWriter
from .base import Place, Transition
from .ui import UIInternalArrow, UIChannelArrow

class PNMLWriter:
	"""Allows exporting a petrinet and a graph in PNML format"""

	def __init__(self, net, graph):
		self.net = net
		self.graph = graph

	def save(self, file):
		"""Write petrinet corresponding to `self.net` and `self.graph` to `file` in PNML format"""
		dictionary = {}
		id = 0
		stream = QXmlStreamWriter(file)
		stream.setAutoFormatting(True)
		stream.writeStartDocument()
		stream.writeStartElement("pnml")
		stream.writeStartElement("net")
		stream.writeAttribute("id", str(id))
		id += 1
		stream.writeAttribute("type", "http://www.pnml.org/version-2009/grammar/pnmlcoremodel")
		stream.writeStartElement("page")
		stream.writeAttribute("id", str(id))
		id += 1
		for enterprise in self.graph.nodes:
			graph = enterprise.obj.graph

			for node in graph.nodes:
				if isinstance(node.obj, Place):
					stream.writeStartElement("place")
				else:
					stream.writeStartElement("transition")
				stream.writeAttribute("id", str(id))
				dictionary[node.obj] = id
				id += 1
				stream.writeAttribute("name", str(node.label.text))
				if isinstance(node.obj, Place):
					stream.writeAttribute("initialMarking", str(node.obj.tokens))
				stream.writeEndElement()
			for arrow in graph.arrows:
				stream.writeStartElement("arc")
				stream.writeAttribute("id", str(id))
				id += 1
				stream.writeAttribute("source", str(dictionary[arrow.source.obj]))
				stream.writeAttribute("target", str(dictionary[arrow.target.obj]))
				stream.writeEndElement()
		for arrow in self.graph.arrows:
			stream.writeStartElement("place")
			shannel = str(id)
			stream.writeAttribute("id", channel)
			id += 1
			stream.writeAttribute("name", str(arrow.source.transition.message))
			stream.writeEndElement()

			stream.writeStartElement("arc")
			stream.writeAttribute("id", str(id))
			id += 1
			stream.writeAttribute("source", str(dictionary[arrow.source.transition]))
			stream.writeAttribute("target", channel)
			stream.writeEndElement()

			stream.writeStartElement("arc")
			stream.writeAttribute("id", str(id))
			id += 1
			stream.writeAttribute("source", channel)
			stream.writeAttribute("target", str(dictionary[arrow.target.transition]))
			stream.writeEndElement()
		stream.writeEndElement()
		stream.writeEndElement()
		stream.writeEndElement()
		stream.writeEndDocument()
		file.close()




from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from ..common import Signal
import string
import json
import os
		
		
KeyCodes = {string.ascii_uppercase[i]: Qt.Key_A + i for i in range(26)}
		

class Tool:
	def __init__(self, data):
		self.name = data["name"]
		self.shape = data["shape"]
		self.text = data["text"]
		self.tooltip = data["tooltip"]
		self.shortcut = KeyCodes[data["shortcut"]]
		
		
class ToolGroup:
	def __init__(self):
		self.name = None
		self.shape = None
		self.tools = []
		

class ToolList:
	def __init__(self):
		self.groups = {}
		self.tools = {}
		
	def load(self, filename):
		with open(filename) as f:
			data = json.load(f)
		
		for info in data["tools"]:
			tool = Tool(info)
			self.tools[tool.name] = tool
			
		for info in data["groups"]:
			group = ToolGroup()
			group.name = info["name"]
			group.shape = info["shape"]
			for name in info["tools"]:
				group.tools.append(self.tools[name])
			self.groups[group.name] = group
			
			
class ToolButton(QToolButton):
	def __init__(self, style, tool):
		super().__init__()
		self.setToolTip(tool.tooltip)

		self.setFixedWidth(80)
		self.setFixedHeight(80)
		self.setCheckable(True)
		self.shape = style.shapes[tool.shape]
		self.text = tool.text
		self.name = tool.name

		self.font = QFont()
		self.font.setPixelSize(16)
		self.textRect = QRectF(0, 55, 80, 20)

		self.clicked.connect(lambda: self.setChecked(True))

	def paintEvent(self, e):
		super().paintEvent(e)
		
		painter = QPainter()
		painter.begin(self)
		
		painter.translate(40, 30)
		self.shape.draw(painter)
		painter.resetTransform()

		painter.setFont(self.font)
		painter.drawText(self.textRect, Qt.AlignCenter, self.text)

		painter.end()
		
		
class ToolBarHeader(QWidget):
	def __init__(self, style, group):
		super().__init__()
		self.setMinimumWidth(40)
		self.setMinimumHeight(40)
		self.shape = style.shapes[group.shape]
		self.orientation = Qt.Vertical
		
	def setOrientation(self, orientation):
		self.orientation = orientation
		self.update()
		
	def paintEvent(self, e):
		super().paintEvent(e)
		
		painter = QPainter()
		painter.begin(self)
		
		painter.save()
		
		if self.orientation == Qt.Horizontal:
			gradient = QLinearGradient(0, 0, self.width(), 0)
		else:
			gradient = QLinearGradient(0, 0, 0, self.height())
		gradient.setColorAt(0, QColor(160, 160, 160))
		gradient.setColorAt(1, QPalette().color(QPalette.Window))
		
		painter.setPen(Qt.NoPen)
		painter.setBrush(gradient)
		painter.drawRect(self.rect())
		
		painter.restore()
		
		painter.translate(self.rect().center())
		self.shape.draw(painter)
		
		painter.end()


class ToolBar(QToolBar):
	"""The menu bar that can be used to select the active tools for left and right mouse buttons"""
	def __init__(self, style):
		super().__init__()
		self.selectionChanged = Signal()
		
		self.setFloatable(False)
		
		self.style = style
		
		self.tools = ToolList()
		self.tools.load(os.path.join(os.path.dirname(__file__), "../data/tools.json"))
		
		self.reset()
		
	def reset(self):
		"""Empty the toolbar"""
		self.clear()
		
		self.groups = {}
		self.buttons = {}
		self.shortcuts = {}
		
	def addGroup(self, name):
		"""Make a group of tools availabe, so that their corresponding tool buttons may be visible in the toolbar."""
		group = self.tools.groups[name]
		
		header = ToolBarHeader(self.style, group)
		self.orientationChanged.connect(header.setOrientation)
		self.addWidget(header)
		
		buttonGroup = QButtonGroup(self)
		buttonGroup.buttonToggled.connect(self.handleToggled)
		for tool in group.tools:
			button = ToolButton(self.style, tool)
			buttonGroup.addButton(button)
			self.addWidget(button)
			self.buttons[tool.name] = button
			self.shortcuts[tool.shortcut] = tool.name
		
		self.groups[group.name] = buttonGroup
		
	def handleToggled(self, button, state):
		if state:
			self.selectionChanged.emit(button.name)

	def handleKey(self, key):
		if key in self.shortcuts:
			self.selectTool(self.shortcuts[key])
			
	def selectTool(self, name):
		"""Select and enable the tool with name `name`"""
		self.buttons[name].setChecked(True)

	def currentTool(self, group):
		"""Return the currently active tool for a specific group, if there is an active too in said group.
		Otherwise, return -1."""
		button = self.groups[group].checkedButton()
		if button:
			return button.name
		return -1



"""This module allows for signalling with callbacks and interconnectable signals.
To do this, the :py:class:`~in_toolset.common.Signal` and :py:class:`~in_toolset.common.SignalListener` classes are used.
Furthermore, it provides the class :py:class:`~in_toolset.common.Property` which allows the creation of object properties with associated signals to signify changes."""

class Signal:
	"""A Signal basically consists of a list of callbacks and associated params that will be called when :py:meth:`~in_toolset.common.Signal.emit` is called."""
	def __init__(self):
		self.callbacks = []
		
	def connect(self, func, *param):
		"""Connect a callback, to be called when when :py:meth:`~in_toolset.common.Signal.emit` is called."""
		self.callbacks.append((func, param))
		
	def disconnect(self, func, *param):
		"""Disconnect or remove a callback so that it will no longer be called when :py:meth:`~in_toolset.common.Signal.emit` is called."""
		self.callbacks.remove((func, param))
		
	def emit(self, *args):
		"""Call all connected callbacks with `args` and their associated `param`"""
		for func, param in self.callbacks:
			func(*args, *param)
			
	__call__ = emit
	
	
class SignalListener:
	"""The SignalListener class groups callbacks and the signals to which they were added to allow for easier deletion/disconnection."""

	def __init__(self):
		self.listeners = []
		
	def connect(self, signal, callback):
		"""Connect `callback` to `signal` and store it in `this`"""
		self.listeners.append((signal, callback))
		signal.connect(callback)
		
	def disconnect(self):
		"""Disconnect all callbacks stored in `this`"""
		for signal, callback in self.listeners:
			signal.disconnect(callback)
		self.listeners = []

			
class Property:
	"""A basic property system that can be used to add properties with getters, setters, and signals to a class.."""
	def __init__(self, signame, default=None):
		self.signame = signame
		self.default = default
		
	def read(self, inst): return inst.__dict__.get(self.name, self.default)
	def write(self, inst, value):
		inst.__dict__[self.name] = value
		inst.__dict__[self.signame].emit()
		
	def __set__(self, instance, value):
		old = self.read(instance)
		if value != old:
			self.write(instance, value)
			
	def __get__(self, instance, owner):
		return self.read(instance)
		
	def __set_name__(self, owner, name):
		self.name = name




from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from ..common import Signal, SignalListener


class SeparatorLine(QFrame):
	def __init__(self):
		super().__init__()
		self.setFrameShape(QFrame.HLine)
		self.setFrameShadow(QFrame.Sunken)


class SettingsWidget(QWidget):
	def __init__(self):
		super().__init__()
		self.setStyleSheet("font-size: 16px")
		
		self.layout = QFormLayout(self)
		
		self.signals = SignalListener()
		
	def connect(self, signal, callback):
		self.signals.connect(signal, callback)
		
	def cleanup(self):
		self.signals.disconnect()
		
	def addSeparator(self):
		self.layout.addRow(SeparatorLine())
		
	def addField(self, label, widget):
		self.layout.addRow(label, widget)
		
	def addWidget(self, widget):
		self.layout.addRow(widget)
		
		
class GeneralSettings(SettingsWidget):
	def __init__(self, net):
		super().__init__()
		
		self.net = net.net
		self.connect(self.net.deadlockChanged, self.updateDeadlock)
		
		self.label = QLabel("No item selected")
		self.label.setAlignment(Qt.AlignCenter)
		
		self.triggerRandom = QPushButton("Trigger random")
		self.triggerRandom.setEnabled(not self.net.deadlock)
		self.triggerRandom.clicked.connect(self.net.triggerRandom)
		
		self.addWidget(self.label)
		self.addWidget(self.triggerRandom)
		
	def setSelection(self, items):
		if len(items) == 0:
			self.label.setText("No item selected")
		elif len(items) == 1:
			self.label.setText("1 item selected")
		else:
			self.label.setText("%i items selected" %len(items))
			
	def updateDeadlock(self):
		self.triggerRandom.setEnabled(not self.net.deadlock)


class PetriScene:
	def __init__(self, style, window):
		self.style = style
		self.window = window
		
		self.toolbar = self.window.toolbar
		self.scene = self.window.scene
		self.view = self.window.view
		self.settings = self.window.settings
		
		self.enterpriseSelected = Signal()
		
		self.signals = SignalListener()
		
	def loadPetriNet(self, net):
		self.scene.clear()
		self.view.resetTransform()
		
		self.net = net
		self.connect(net.graph.nodes.added, self.addNode)
		self.connect(net.graph.arrows.added, self.addArrow)
		self.connect(net.graph.looseArrows.added, self.addLooseArrow)
		
		for node in net.graph.nodes:
			self.addNode(node)
		for arrow in net.graph.arrows:
			self.addArrow(arrow)
		for arrow in net.graph.looseArrows:
			self.addLooseArrow(arrow)
		
		self.controller = self.createController()
		self.scene.setController(self.controller)
		
		self.connect(self.scene.selectionChanged, self.updateSelection)
		
		self.view.setHandDrag(False)
		
		self.toolbar.reset()
		self.toolbar.addGroup("common")
		self.toolbar.selectTool("selection")
		self.registerTools(self.toolbar)
		self.connect(self.toolbar.selectionChanged, self.updateTool)
		
		self.generalSettings = self.createGeneralSettings()
		self.settings.setWidget(self.generalSettings)
		
	def cleanup(self):
		self.signals.disconnect()
		
		self.generalSettings.cleanup()
		if self.settings.widget() != self.generalSettings:
			self.settings.widget().cleanup()
		self.scene.cleanup()
		
	def updateSelection(self):
		if self.settings.widget() != self.generalSettings:
			self.settings.widget().cleanup()

		items = self.scene.selectedItems()
		widget = self.createSettingsWidget(items)
		if widget:
			self.settings.setWidget(widget)
		else:
			self.generalSettings.setSelection(items)
			self.settings.setWidget(self.generalSettings)
		
	def updateTool(self, tool):
		if tool == "selection":
			self.view.setHandDrag(False)
		elif tool == "hand":
			self.view.setHandDrag(True)
		
	def connect(self, signal, callback):
		self.signals.connect(signal, callback)
		
	def registerTools(self, toolbar): pass
	
	def addNode(self, node): pass
	def addArrow(self, arrow): pass
	def addLooseArrow(self, arrow): pass
	
	def createGeneralSettings(self): return GeneralSettings(self.net)
	def createSettingsWidget(self, items): return None
	
	def createController(self): raise NotImplementedError





from ..common import Signal, Property
from .base import *
from .ui import *
from .pnml import *
from PyQt5.QtCore import QFile
from PyQt5.QtCore import QIODevice
import json


class ProjectReader:
	"""Loads a project file (in our own json-based file format) corresponding to an inet into the program."""
	def load(self, data):
		self.loadPlaces(data)
		self.loadTransitions(data)
		self.loadEnterprises(data)

		industry = UIPetriNet()
		for place in self.places:
			industry.net.places.add(place)
		for trans in self.transitions:
			industry.net.transitions.add(trans)
		self.loadGraph(industry.graph, data["industry"], True)
		return industry

	def loadPlaces(self, data):
		self.places = []
		for info in data["places"]:
			place = Place()
			place.tokens = info
			self.places.append(place)

	def loadTransitions(self, data):
		self.transitions = []
		for info in data["transitions"]:
			transition = UITransition()
			transition.type = info["type"]
			transition.message = info["message"]
			self.transitions.append(transition)

	def loadEnterprises(self, data):
		self.enterprises = []
		for info in data["enterprises"]:
			enterprise = UIPetriNet()
			self.loadPetriNet(enterprise.net, info["net"])
			self.loadGraph(enterprise.graph, info["graph"], False)
			self.enterprises.append(enterprise)

	def loadPetriNet(self, net, data):
		for place in data["places"]:
			net.places.add(self.places[place])
		for transition in data["transitions"]:
			net.transitions.add(self.transitions[transition])

	def loadGraph(self, graph, data, industry):
		for info in data["nodes"]:
			obj = self.parseObjectRef(info)
			node = UINode(obj)
			self.loadPosition(node, info["pos"])
			self.loadLabel(node.label, info["label"])
			graph.nodes.add(node)

		for info in data["looseArrows"]:
			node = graph.nodes[info["node"]]
			transition = self.transitions[info["transition"]]

			arrow = UILooseArrow(node, transition)
			arrow.angle = info["angle"]
			self.loadLabel(arrow.label, info["label"])

			graph.looseArrows.add(arrow)

		for info in data["arrows"]:
			if industry:
				source = graph.looseArrows[info["source"]]
				target = graph.looseArrows[info["target"]]
				channel = self.places[info["channel"]]

				arrow = UIChannelArrow(source, target, channel)
			else:
				source = graph.nodes[info["source"]]
				target = graph.nodes[info["target"]]

				arrow = UIInternalArrow(source, target)

			arrow.curve = info["curve"]

			graph.arrows.add(arrow)

	def loadPosition(self, obj, data):
		obj.x = data[0]
		obj.y = data[1]

	def loadLabel(self, label, data):
		label.text = data["text"]
		label.angle = data["angle"]
		label.distance = data["distance"]

	def parseObjectRef(self, data):
		type = data["type"]
		id = data["id"]
		if type == "place": return self.places[id]
		if type == "transition": return self.transitions[id]
		if type == "enterprise": return self.enterprises[id]
		raise ValueError("Invalid object type: %s" %type)


class ProjectWriter:
	"""Writes a project to a project file (in our own json-based file format)."""
	def save(self, industry):
		data = {}
		data["places"] = self.savePlaces(industry)
		data["transitions"] = self.saveTransitions(industry)
		data["enterprises"] = self.saveEnterprises(industry)
		data["industry"] = self.saveGraph(industry.graph)
		return data

	def savePlaces(self, industry):
		places = []
		for place in industry.net.places:
			place.id = len(places)
			places.append(place.tokens)
		return places

	def saveTransitions(self, industry):
		transitions = []
		for trans in industry.net.transitions:
			trans.id = len(transitions)
			transitions.append({
				"type": trans.type,
				"message": trans.message
			})
		return transitions

	def saveEnterprises(self, industry):
		enterprises = []
		for node in industry.graph.nodes:
			node.obj.id = len(enterprises)
			enterprise = {
				"net": self.savePetriNet(node.obj.net),
				"graph": self.saveGraph(node.obj.graph)
			}
			enterprises.append(enterprise)
		return enterprises

	def savePetriNet(self, net):
		places = []
		for place in net.places:
			places.append(place.id)
		transitions = []
		for trans in net.transitions:
			transitions.append(trans.id)
		return {
			"places": places,
			"transitions": transitions
		}

	def saveGraph(self, graph):
		nodes = []
		for node in graph.nodes:
			node.id = len(nodes)
			info = self.saveObjectRef(node.obj)
			info["pos"] = self.savePosition(node)
			info["label"] = self.saveLabel(node.label)
			nodes.append(info)

		messages = []
		for arrow in graph.looseArrows.objects:
			if arrow.transition.active:
				arrow.id = len(messages)
				messages.append({
					"node": arrow.node.id,
					"transition": arrow.transition.id,
					"angle": arrow.angle,
					"label": self.saveLabel(arrow.label)
				})

		arrows = []
		for arrow in graph.arrows:
			info = {
				"source": arrow.source.id,
				"target": arrow.target.id,
				"curve": arrow.curve
			}
			if isinstance(arrow, UIChannelArrow):
				info["channel"] = arrow.channel.id
			arrows.append(info)

		return {
			"nodes": nodes,
			"arrows": arrows,
			"looseArrows": messages
		}

	def savePosition(self, obj):
		return [obj.x, obj.y]

	def saveLabel(self, label):
		return {
			"text": label.text,
			"angle": label.angle,
			"distance": label.distance
		}

	def saveObjectRef(self, obj):
		return {
			"type": self.getObjectType(obj),
			"id": obj.id
		}

	def getObjectType(self, obj):
		if isinstance(obj, Place): return "place"
		if isinstance(obj, Transition): return "transition"
		if isinstance(obj, UIPetriNet): return "enterprise"
		raise ValueError("This should never happen")


class Project:
	filename = Property("filenameChanged")
	unsaved = Property("unsavedChanged", False)

	def __init__(self):
		self.filenameChanged = Signal()
		self.unsavedChanged = Signal()

		self.industry = UIPetriNet()
		self.industry.changed.connect(self.setUnsaved)

		self.reader = ProjectReader()
		self.writer = ProjectWriter()

		self.exportname = ""

	def setFilename(self, filename): self.filename = filename
	def setExportname(self, exportname): self.exportname = exportname
	def setUnsaved(self, unsaved=True): self.unsaved = unsaved

	def load(self, filename):
		with open(filename) as f:
			data = json.load(f)

		self.industry = self.reader.load(data)
		self.industry.changed.connect(self.setUnsaved)

		self.setFilename(filename)
		self.setUnsaved(False)

	def save(self, filename):
		data = self.writer.save(self.industry)
		with open(filename, "w") as f:
			json.dump(data, f)

		self.setFilename(filename)
		self.setUnsaved(False)

	def export(self, exportname):
		writer = PNMLWriter(self.industry.net, self.industry.graph)
		file = QFile(exportname)
		if file.open(QIODevice.WriteOnly):
			stream = writer.save(file)



"""This library adds the ability to store data necessary for a visual representation of an industry net."""

from .base import Object, ObjectList, PetriNet, Transition, Place
from ..common import Signal, Property
import math


class TransitionType:
	"""An enum for the different possible transition types"""
	INTERNAL = 0 #: An internal transition "stays" within a single enterprise net and is not connected to other nets
	INPUT = 1    #: An input transition can receive messages from another enterprise if connected on the industry level
	OUTPUT = 2   #: An output transition can send messages to another enterprise if enabled and connected on the industry level

class UITransition(Transition):
	type = Property("typeChanged", TransitionType.INTERNAL)
	message = Property("messageChanged", "")
	channel = Property("channelChanged", None)

	def __init__(self):
		super().__init__()
		self.typeChanged = Signal()
		self.typeChanged.connect(self.changed)
		self.messageChanged = Signal()
		self.messageChanged.connect(self.changed)
		self.channelChanged = Signal()
		self.channelChanged.connect(self.changed)

	def setType(self, type): self.type = type
	def setMessage(self, message): self.message = message
	def setChannel(self, channel): self.channel = channel


# Subclassed by: UILabel, UINode, UILooseArrow
class UIObject(Object):
	x = Property("positionChanged", 0)
	y = Property("positionChanged", 0)

	def __init__(self):
		super().__init__()
		self.positionChanged = Signal()
		self.positionChanged.connect(self.changed)

	def move(self, x, y):
		self.x = x
		self.y = y


class UILabel(UIObject):
	text = Property("textChanged", "")
	angle = Property("angleChanged", math.pi / 2)
	distance = Property("distanceChanged", 35)

	def __init__(self):
		super().__init__()
		self.textChanged = Signal()
		self.angleChanged = Signal()
		self.distanceChanged = Signal()

		self.textChanged.connect(self.changed)
		self.angleChanged.connect(self.changed)
		self.distanceChanged.connect(self.changed)

	def setText(self, text): self.text = text
	def setAngle(self, angle): self.angle = angle
	def setDistance(self, dist): self.distance = dist


# Holds UIPetriNet, UITransition or UIPlace
class UINode(UIObject):
	def __init__(self, obj):
		super().__init__()
		self.obj = obj
		self.obj.changed.connect(self.changed)
		self.obj.deleted.connect(self.delete)

		self.label = UILabel()
		self.deleted.connect(self.label.delete)
		self.deleted.connect(self.obj.delete)

		self.positionChanged.connect(self.updateLabel)

	def updateLabel(self):
		self.label.move(self.x, self.y)


# Arrow between instances of UILooseArrow or UINode
class UIArrow(Object):
	curve = Property("curveChanged", 0)

	def __init__(self, source, target):
		super().__init__()
		self.source = source
		self.target = target

		source.deleted.connect(self.delete)
		target.deleted.connect(self.delete)

		self.curveChanged = Signal()
		self.curveChanged.connect(self.changed)

	def setCurve(self, curve): self.curve = curve


# Arrow between instances of UINode (enterprise)
class UIInternalArrow(UIArrow):
	def __init__(self, source, target):
		super().__init__(source, target)

		source.obj.connect(target.obj)

		self.deleted.connect(self.disconnect)

	def disconnect(self):
		self.source.obj.disconnect(self.target.obj)


# Arrow between instances of UILooseArrow (industry)
class UIChannelArrow(UIArrow):
	def __init__(self, source, target, channel):
		super().__init__(source, target)

		self.channel = channel
		self.channel.deleted.connect(self.delete)
		self.deleted.connect(self.channel.delete)
		self.deleted.connect(self.unregister)

		source.transition.setChannel(self.channel)
		target.transition.setChannel(self.channel)

		source.transition.connect(self.channel)
		self.channel.connect(target.transition)

	def unregister(self):
		self.source.transition.setChannel(None)
		self.target.transition.setChannel(None)


# Message arrow around UINode
# Belongs to a UITransition
class UILooseArrow(UIObject):
	angle = Property("angleChanged", math.pi)

	def __init__(self, node, transition):
		super().__init__()
		self.angleChanged = Signal()
		self.angleChanged.connect(self.updatePos)

		self.node = node
		self.node.positionChanged.connect(self.updatePos)
		self.node.deleted.connect(self.delete)

		self.transition = transition
		self.transition.messageChanged.connect(self.updateText)
		self.transition.typeChanged.connect(self.updateType)
		self.transition.deleted.connect(self.delete)

		self.label = UILabel()
		self.label.distance = 15
		self.label.textChanged.connect(self.updateMessage)
		self.deleted.connect(self.label.delete)
		self.restored.connect(self.label.restore)

		self.deleted.connect(self.unregister)

		self.updatePos()
		self.updateText()
		self.updateType()

	def setAngle(self, angle): self.angle = angle

	def unregister(self):
		self.transition.setType(TransitionType.INTERNAL)

	def updateType(self):
		if self.transition.type == TransitionType.INTERNAL:
			self.delete()
		else:
			self.restore()

	def updatePos(self):
		x = self.node.x + math.cos(self.angle) * 40
		y = self.node.y + math.sin(self.angle) * 40
		self.move(x, y)
		self.label.move(x, y)

	def updateMessage(self):
		self.transition.setMessage(self.label.text)

	def updateText(self):
		self.label.setText(self.transition.message)


# Graph with UINode, UIArrow and UILooseArrow objects
class UIGraph(Object):
	def __init__(self):
		super().__init__()

		self.nodes = ObjectList()
		self.nodes.changed.connect(self.changed)

		self.arrows = ObjectList()
		self.arrows.changed.connect(self.changed)

		self.looseArrows = ObjectList()
		self.looseArrows.changed.connect(self.changed)


# Petri net with graph
class UIPetriNet(Object):
	def __init__(self):
		super().__init__()

		self.net = PetriNet()
		self.net.changed.connect(self.changed)

		self.graph = UIGraph()
		self.graph.changed.connect(self.changed)



"""The config module allows for configuration of the behaviour of in-toolset,
using a text file with key-value items as storage for the settings"""

import os

class Config:
	"""Manages basic configuration of the graphical interface as a list of key-value pairs"""
	types = {
		"ui.max_label_size": int,
		"ui.keyboard_scroll_speed": float
	}
	
	def __init__(self, filename):
		self.load(filename)
		
	def get(self, field):
		"""Get the value of `field`"""
		return self.settings[field]
	
	def set(self, field, value):
		"""Set the value of `field` to `value`"""
		if field not in self.types:
			raise ValueError("Unknown setting: %s" %field)
		self.settings[field] = self.types[field](value)
		
	def load(self, filename):
		"""Load the key-value pairs from the file `filename` into `self`"""
		self.settings = {}
		with open(filename) as f:
			for index, line in enumerate(f):
				line = line.strip()
				if line:
					if "=" in line:
						field, value = line.split("=", 1)
						self.set(field.strip(), value.strip())
					else:
						raise ValueError("Syntax error at line %i" %index)


config = Config(os.path.join(os.path.dirname(__file__), "data/config.txt"))

def get(field):
	"""Get the current value of the configuration field `field`"""
	return config.get(field)




from PyQt5.QtWidgets import *


class Action(QAction):
	def __init__(self, text, shortcut, checkable=False):
		super().__init__(text)
		self.setShortcut(shortcut)
		self.setCheckable(checkable)


class FileMenu(QMenu):
	def __init__(self):
		super().__init__("File")

		self.new = Action("New", "Ctrl+N")
		self.open = Action("Open", "Ctrl+O")
		self.save = Action("Save", "Ctrl+S")
		self.saveAs = Action("Save as", "Ctrl+Shift+S")
		self.export = Action("Export", "Ctrl+E")
		self.exportAs = Action("Export as", "Ctrl+Shift+E")
		self.quit = Action("Quit", "Ctrl+Q")

		self.addAction(self.new)
		self.addAction(self.open)
		self.addAction(self.save)
		self.addAction(self.saveAs)
		self.addAction(self.export)
		self.addAction(self.exportAs)
		self.addAction(self.quit)


class EditMenu(QMenu):
	def __init__(self):
		super().__init__("Edit")

		self.selectAll = Action("Select all", "Ctrl+A")
		self.setInitialMarking = Action("Set Initial Marking", "Ctrl+M")

		self.addAction(self.selectAll)
		self.addAction(self.setInitialMarking)


class ViewMenu(QMenu):
	def __init__(self):
		super().__init__("View")

		self.showGrid = Action("Show grid", "Ctrl+1", True)
		self.showGrid.setChecked(True)
		self.resetCamera = Action("Reset camera", "Ctrl+R")
		self.editIndustry = Action("Go to industry net", "Ctrl+I")

		self.addAction(self.showGrid)
		self.addSeparator()
		self.addAction(self.resetCamera)
		self.addAction(self.editIndustry)


class MenuBar(QMenuBar):
	def __init__(self):
		super().__init__()

		self.file = FileMenu()
		self.edit = EditMenu()
		self.view = ViewMenu()

		self.addMenu(self.file)
		self.addMenu(self.edit)
		self.addMenu(self.view)




from PyQt5.QtWidgets import *
from ..model.project import Project
from ..model.ui import UIPetriNet
from .industry import IndustryScene
from .enterprise import EnterpriseScene
from .window import MainWindow
from .view import Style
import sys
import os


class Application:
	"""The main class, corresponding to one instance of the graphical toolkit.
	It manages windows and allows for basic managing of projects."""

	def start(self):
		"""Start the graphical interface and the program itself"""
		self.app = QApplication(sys.argv)
		
		style = Style()
		style.load(os.path.join(os.path.dirname(__file__), "../data/style.json"))
		
		self.window = MainWindow(style)
		self.window.newProject.connect(self.createProject)
		self.window.loadProject.connect(self.createProject)
		self.window.enterpriseSelected.connect(self.switchToScene)
		self.window.show()
		
		self.industryScene = IndustryScene(style, self.window)
		self.industryScene.enterpriseSelected.connect(self.switchToScene)

		self.enterpriseScene = EnterpriseScene(style, self.window)
		
		self.currentScene = None
		
		self.createProject()
		
		self.app.exec()
		
	def createProject(self, filename=None):
		"""Create a new project or load an existing one"""

		self.project = Project()
		if filename:
			try:
				self.project.load(filename)
			except:
				import traceback
				traceback.print_exc()
				
				text = "An error occurred while loading this file (it may be corrupted)."
				QMessageBox.warning(self.window, "Error", text)
				return
			
		self.industry = self.project.industry
			
		self.window.setProject(self.project)
		self.switchToScene(self.industry)
		
	def switchToScene(self, object):
		if self.currentScene:
			self.currentScene.cleanup()
		
		if object == self.industry:
			self.currentScene = self.industryScene
			self.currentScene.load(self.industry)
		else:
			self.currentScene = self.enterpriseScene
			self.currentScene.load(self.industry, object)




import sys

def main():
	"""Start the graphical editor with a new, blank, industry net."""
	version = sys.version_info
	if version < (3, 6):
		version = "%i.%i.%i" %(version.major, version.minor, version.micro)
		msg = "Your Python version is old (%s). Please upgrade to 3.6 or higher." %version
		raise RuntimeError(msg)

	from .ui.app import Application

	app = Application()
	app.start()




from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from .view import EditorScene, EditorView
from .tools import ToolBar
from .menu import MenuBar
from . import settings
from ..common import Signal
from ..model.project import Project
import os


class IndustryItem(QListWidgetItem):
	def __init__(self, industry):
		super().__init__("Industry")
		self.obj = industry


class EnterpriseItem(QListWidgetItem):
	def __init__(self, node):
		super().__init__()

		self.obj = node
		self.obj.label.textChanged.connect(self.updateText)
		self.updateText()

	def updateText(self):
		name = self.obj.label.text
		if not name:
			name = "Enterprise"
		self.setText(name)


class NetListWidget(QWidget):
	def __init__(self):
		super().__init__()
		self.enterpriseSelected = Signal()
		layout = QVBoxLayout()

		self.editIndustry = QPushButton("Go to industry net")
		self.editIndustry.clicked.connect(self.handleEditIndustry)
		layout.addWidget(self.editIndustry)

		self.listWidget = QListWidget()
		layout.addWidget(self.listWidget)

		self.setLayout(layout)
		self.show()

	def setProject(self, project):
		self.listWidget.clear()

		project.industry.graph.nodes.added.connect(self.addEnterprise)
		project.industry.graph.nodes.removed.connect(self.removeEnterprise)

		self.listWidget.addItem(IndustryItem(project.industry))
		for enterprise in project.industry.graph.nodes:
			self.addEnterprise(enterprise)

		self.listWidget.itemActivated.connect(self.handleItemActivated)

	def addEnterprise(self, enterprise):
		item = EnterpriseItem(enterprise)
		self.listWidget.addItem(item)

	def removeEnterprise(self, enterprise):
		for i in range(self.listWidget.count()):
			if self.listWidget.item(i).obj == enterprise:
				self.listWidget.takeItem(i)
				return

	def handleEditIndustry(self):
		industry = self.listWidget.item(0)
		self.listWidget.setCurrentItem(industry)
		self.enterpriseSelected.emit(industry.obj)

	def handleItemActivated(self, item):
		self.enterpriseSelected.emit(item.obj)


class MainWindow(QMainWindow):
	def __init__(self, style):
		super().__init__()
		self.newProject = Signal()
		self.loadProject = Signal()
		self.enterpriseSelected = Signal()

		self.setContextMenuPolicy(Qt.PreventContextMenu)

		self.resize(1080, 720)

		self.toolbar = ToolBar(style)
		self.addToolBar(Qt.LeftToolBarArea, self.toolbar)

		self.scene = EditorScene()
		self.view = EditorView(self.scene)
		self.setCentralWidget(self.view)

		self.settings = QDockWidget("Settings")
		self.settings.setFixedWidth(200)
		self.settings.setFeatures(QDockWidget.DockWidgetMovable)
		self.settings.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
		self.addDockWidget(Qt.RightDockWidgetArea, self.settings)

		self.nets = NetListWidget()
		self.nets.enterpriseSelected.connect(self.enterpriseSelected)
		netsDock = QDockWidget("Nets")
		netsDock.setFixedWidth(200)
		netsDock.setFeatures(QDockWidget.DockWidgetMovable)
		netsDock.setAllowedAreas(Qt.LeftDockWidgetArea | Qt.RightDockWidgetArea)
		netsDock.setWidget(self.nets)
		self.addDockWidget(Qt.RightDockWidgetArea, netsDock)

		menuBar = MenuBar()
		menuBar.file.new.triggered.connect(self.handleNew)
		menuBar.file.open.triggered.connect(self.handleOpen)
		menuBar.file.save.triggered.connect(self.handleSave)
		menuBar.file.saveAs.triggered.connect(self.handleSaveAs)
		menuBar.file.export.triggered.connect(self.handleExport)
		menuBar.file.exportAs.triggered.connect(self.handleExportAs)
		menuBar.file.quit.triggered.connect(self.close)
		menuBar.edit.selectAll.triggered.connect(self.scene.selectAll)
		menuBar.edit.setInitialMarking.triggered.connect(self.handleSetInitialMarking)
		menuBar.view.showGrid.toggled.connect(self.scene.setGridEnabled)
		menuBar.view.resetCamera.triggered.connect(self.view.resetTransform)
		menuBar.view.editIndustry.triggered.connect(self.selectIndustry)
		self.setMenuBar(menuBar)

	def setProject(self, project):
		self.nets.setProject(project)

		self.project = project
		self.project.filenameChanged.connect(self.updateWindowTitle)
		self.project.unsavedChanged.connect(self.updateWindowTitle)

		self.updateWindowTitle()

	def handleSetInitialMarking(self):
		self.project.industry.net.setInitialMarking()

	def selectIndustry(self):
		self.enterpriseSelected.emit(self.project.industry)

	def closeEvent(self, e):
		if self.checkUnsaved():
			e.accept()
		else:
			e.ignore()

	def keyPressEvent(self, e):
		self.toolbar.handleKey(e.key())

	def updateWindowTitle(self):
		name = self.project.filename
		if name is None:
			name = "untitled"
		self.setWindowTitle("Petri - %s%s" %(name, "*" * self.project.unsaved))

	def checkUnsaved(self):
		if self.project.unsaved:
			msg = "This model has unsaved changes. Do you want to save them?"
			buttons = QMessageBox.Save | QMessageBox.Discard | QMessageBox.Cancel
			result = QMessageBox.question(self, "Save changes?", msg, buttons)
			if result == QMessageBox.Save: return self.handleSave()
			elif result == QMessageBox.Discard: return True
			else: return False
		return True

	def handleNew(self):
		if self.checkUnsaved():
			self.newProject.emit()
			return True
		return False

	def handleOpen(self):
		if not self.checkUnsaved():
			return False

		filename, filter = QFileDialog.getOpenFileName(
			self, "Load model", settings.getLastPath(),
			"Workflow model (*.flow);;All files (*.*)"
		)
		if not filename:
			return False

		settings.setLastPath(os.path.dirname(filename))
		self.loadProject.emit(filename)
		return True

	def handleSave(self):
		if not self.project.filename:
			return self.handleSaveAs()
		self.project.save(self.project.filename)
		return True

	def handleSaveAs(self):
		filename, filter = QFileDialog.getSaveFileName(
			self,
			"Save model",
			settings.getLastPath(),
			"Workflow model (*.flow);;All files (*.*)",
			"Workflow model (*.flow)",
			QFileDialog.DontUseNativeDialog
		)
		if not filename:
			return False
		if not '.' in filename:
			filename += ".flow"

		settings.setLastPath(os.path.dirname(filename))
		self.project.save(filename)
		return True

	def handleExport(self):
		if not self.project.exportname:
			return self.handleExportAs()
		self.project.export(self.project.exportname)
		return True

	def handleExportAs(self):
		exportname, filter = QFileDialog.getSaveFileName(
			self,
			"Export model",
			settings.getLastPath(),
			"PNML (*.pnml);;All files (*.*)",
			"PNML (*.pnml)",
			QFileDialog.DontUseNativeDialog
		)
		if not exportname:
			return False
		if not '.' in exportname:
			exportname += ".pnml"

		self.project.export(exportname)
		self.project.exportname = exportname
		return True




from PyQt5.QtWidgets import *
from PyQt5.QtGui import *
from PyQt5.QtCore import *
from ..model.ui import *
from .common import *
from .scene import *
from .. import config as config
import math


class TransitionFilter:
	def __init__(self, item):
		self.base = NodeFilter(item)
		self.item = item

	def applyToPen(self, pen):
		self.base.applyToPen(pen)

	def applyToBrush(self, brush):
		if self.item.node.obj.enabled:
			brush.setColor(QColor("#c0f6b3"))
		self.base.applyToBrush(brush)


class PlaceItem(NodeItem):
	def __init__(self, scene, style, node):
		super().__init__(scene, style.shapes["place"], node)
		self.connect(self.node.obj.tokensChanged, self.update)

		self.font = QFont()
		self.font.setPixelSize(16)

	def paint(self, painter, option, widget):
		super().paint(painter, option, widget)

		if self.node.obj.tokens != 0:
			text = str(self.node.obj.tokens)
			painter.setFont(self.font)
			painter.drawText(self.shp.rect, Qt.AlignCenter, text)


class TransitionItem(NodeItem):
	def __init__(self, scene, style, node):
		super().__init__(scene, style.shapes["transition"], node)
		self.connect(self.node.obj.enabledChanged, self.update)
		self.connect(self.node.obj.triggered, self.flash)
		self.filter = TransitionFilter(self)


class TemporaryPlace(NodeBase):
	"""A placeholder place to be displayed on the canvas during placement of a place, to show the place of the place currently being placed."""
	def __init__(self, scene, style):
		super().__init__(scene, style.shapes["place"])


class TemporaryTransition(NodeBase):
	"""A placeholder transition to be displayed on the canvas during placement of a transition, to show the place of the transition currently being placed."""
	def __init__(self, scene, style):
		super().__init__(scene, style.shapes["transition"])


class TemporaryArrow(ArrowBase):
	"""A placeholder arrow to be displayed on the canvas during placement of a arrow (while it is being dragged), to show the orientation and place of the arrow currently being placed."""
	def __init__(self, scene, source):
		super().__init__(scene)

		self.source = source

	def setTarget(self, x, y):
		self.setPoints(self.source.x(), self.source.y(), x, y)

	def drag(self, param):
		self.setTarget(param.mouse.x(), param.mouse.y())


class EnterpriseController:
	def __init__(self, style, window, industry, enode):
		self.style = style

		self.window = window
		self.toolbar = window.toolbar
		self.scene = window.scene

		self.industry = industry
		self.enterprise = enode.obj
		self.enode = enode

	def startPlacement(self, pos):
		type = self.toolbar.currentTool("enterprise")
		if type == "place":
			self.scene.setHoverEnabled(False)
			item = TemporaryPlace(self.scene, self.style)
			item.setPos(alignToGrid(pos))
			return item
		elif type == "transition":
			self.scene.setHoverEnabled(False)
			item = TemporaryTransition(self.scene, self.style)
			item.setPos(alignToGrid(pos))
			return item
		elif type == "arrow":
			source = self.scene.findItem(pos, NodeItem)
			if source:
				item = TemporaryArrow(self.scene, source)
				item.setTarget(pos.x(), pos.y())
				return item

	def finishPlacement(self, pos, item):
		if isinstance(item, NodeBase):
			pos = alignToGrid(pos)
			x, y = pos.x(), pos.y()

			if not item.invalid:
				if isinstance(item, TemporaryPlace):
					obj = Place()
					self.enterprise.net.places.add(obj)
					self.industry.net.places.add(obj)

					node = UINode(obj)
					node.move(x, y)
					self.enterprise.graph.nodes.add(node)

				else:
					obj = UITransition()
					self.enterprise.net.transitions.add(obj)
					self.industry.net.transitions.add(obj)

					node = UINode(obj)
					node.move(x, y)
					self.enterprise.graph.nodes.add(node)

					arrow = UILooseArrow(node, obj)
					arrow.delete()
					self.enterprise.graph.looseArrows.add(arrow)

					arrow = UILooseArrow(self.enode, obj)
					arrow.delete()
					self.industry.graph.looseArrows.add(arrow)

		elif isinstance(item, TemporaryArrow):
			source = item.source
			target = self.scene.findItem(pos, NodeItem)
			if target and target != source:
				if self.checkConnection(source, target):
					arrow = UIInternalArrow(source.node, target.node)
					self.enterprise.graph.arrows.add(arrow)

		self.scene.setHoverEnabled(True)

	def checkConnection(self, source, target):
		if isinstance(source, PlaceItem) and isinstance(target, PlaceItem):
			QMessageBox.warning(
				self.window, "Invalid connection",
				"Places may not be connected to other places."
			)
			return False

		if isinstance(source, TransitionItem) and isinstance(target, TransitionItem):
			QMessageBox.warning(
				self.window, "Invalid connection",
				"Transitions may not be connected to other transitions."
			)
			return False

		if target.node.obj in source.node.obj.postset:
			QMessageBox.warning(
				self.window, "Duplicate arrow",
				"Only one arrow can be placed from one node to another."
			)
			return False

		return True


class PlaceSettings(SettingsWidget):
	def __init__(self, node):
		super().__init__()
		self.node = node
		self.connect(self.node.positionChanged, self.updatePos)
		self.connect(self.node.label.textChanged, self.updateLabel)

		self.obj = node.obj
		self.connect(self.obj.tokensChanged, self.updateTokens)

		self.setStyleSheet("font-size: 16px")

		self.x = QLabel("%i" %(self.node.x / GRID_SIZE))
		self.x.setAlignment(Qt.AlignRight)
		self.y = QLabel("%i" %(self.node.y / GRID_SIZE))
		self.y.setAlignment(Qt.AlignRight)
		self.label = QLineEdit(self.node.label.text)
		self.label.setMaxLength(config.get("ui.max_label_size"))
		self.label.textEdited.connect(self.node.label.setText)
		self.tokens = QSpinBox()
		self.tokens.setRange(0, 999)
		self.tokens.setValue(self.obj.tokens)
		self.tokens.valueChanged.connect(self.obj.setTokens)

		self.addField("X:", self.x)
		self.addField("Y:", self.y)
		self.addField("Label:", self.label)
		self.addField("Tokens:", self.tokens)

	def updatePos(self):
		self.x.setText("%i" %(self.node.x / GRID_SIZE))
		self.y.setText("%i" %(self.node.y / GRID_SIZE))

	def updateLabel(self):
		self.label.setText(self.node.label.text)

	def updateTokens(self):
		self.tokens.setValue(self.obj.tokens)


class TransitionSettings(SettingsWidget):
	def __init__(self, node):
		super().__init__()

		self.node = node
		self.obj = node.obj

		self.connect(self.node.positionChanged, self.updatePos)
		self.connect(self.node.label.textChanged, self.updateLabel)
		self.connect(self.obj.enabledChanged, self.updateEnabled)
		self.connect(self.obj.typeChanged, self.updateType)
		self.connect(self.obj.messageChanged, self.updateMessage)
		self.connect(self.obj.channelChanged, self.updateChannel)

		self.setStyleSheet("font-size: 16px")

		self.x = QLabel("%i" %(self.node.x / GRID_SIZE))
		self.x.setAlignment(Qt.AlignRight)
		self.y = QLabel("%i" %(self.node.y / GRID_SIZE))
		self.y.setAlignment(Qt.AlignRight)

		self.label = QLineEdit(self.node.label.text)
		self.label.setMaxLength(config.get("ui.max_label_size"))
		self.label.textEdited.connect(self.node.label.setText)

		self.trigger = QPushButton("Trigger")
		self.trigger.setEnabled(self.obj.enabled)
		self.trigger.clicked.connect(self.obj.trigger)

		self.type = QComboBox()
		self.type.addItems(["Internal", "Input", "Output"])
		self.type.setCurrentIndex(self.obj.type)
		self.type.currentIndexChanged.connect(self.obj.setType)

		self.message = QLineEdit(self.obj.message)
		self.message.setMaxLength(config.get("ui.max_label_size"))
		self.message.textEdited.connect(self.obj.setMessage)
		self.updateChannel()

		self.addField("X:", self.x)
		self.addField("Y:", self.y)
		self.addField("Label:", self.label)
		self.addSeparator()
		self.addField("Type:", self.type)
		self.addField("Message Type:", self.message)
		self.addSeparator()
		self.addWidget(self.trigger)

	def updatePos(self):
		self.x.setText("%i" %(self.node.x / GRID_SIZE))
		self.y.setText("%i" %(self.node.y / GRID_SIZE))

	def updateLabel(self):
		self.label.setText(self.node.label.text)

	def updateEnabled(self):
		self.trigger.setEnabled(self.obj.enabled)

	def updateType(self):
		self.type.setCurrentIndex(self.obj.type)
		self.message.setEnabled(
			self.obj.type != TransitionType.INTERNAL and not self.obj.channel
		)

	def updateMessage(self):
		self.message.setText(self.obj.message)

	def updateChannel(self):
		self.type.setEnabled(not self.obj.channel)
		self.message.setEnabled(
			self.obj.type != TransitionType.INTERNAL and not self.obj.channel
		)


class EnterpriseScene(PetriScene):
	def load(self, industry, node):
		self.industry = industry
		self.enterprise = node.obj
		self.enode = node

		self.loadPetriNet(node.obj)

	def registerTools(self, toolbar):
		toolbar.addGroup("enterprise")
		toolbar.selectTool("place")

	def createController(self):
		return EnterpriseController(self.style, self.window, self.industry, self.enode)

	def addNode(self, node):
		if isinstance(node.obj, Place):
			item = PlaceItem(self.scene, self.style, node)
		else:
			item = TransitionItem(self.scene, self.style, node)
		self.scene.addItem(item)

	def addArrow(self, arrow):
		item = ArrowItem(self.scene, arrow)
		item.setDistance(35)
		self.scene.addItem(item)

	def addLooseArrow(self, arrow):
		item = LooseArrowItem(self.scene, self.style, arrow)
		self.scene.addItem(item)

	def createSettingsWidget(self, items):
		filtered = [i for i in items if isinstance(i, NodeItem)]
		if len(filtered) == 1:
			item = filtered[0]
			if isinstance(item, PlaceItem):
				return PlaceSettings(item.node)
			return TransitionSettings(item.node)




from ..common import Signal, Property
import random


class Object:
	"""A basic object class, objects can be marked as deleted and have :py:class:`~in_toolset.common.Signal`-typed members to signify deletion or changes"""
	active = Property("statusChanged", True)

	def __init__(self):
		self.changed = Signal()
		self.deleted = Signal()
		self.restored = Signal()

		self.statusChanged = Signal()
		self.statusChanged.connect(self.changed)
		self.statusChanged.connect(self.notifyState)

		self.id = None

	def notifyState(self):
		if self.active:
			self.restored.emit()
		else:
			self.deleted.emit()

	def delete(self):
		"""Mark the object as deleted"""
		self.active = False

	def restore(self):
		"""Mark the object as active"""
		self.active = True


class ObjectList:
	"""A list of objects of the :py:class:`~in_toolset.model.base.Object` class.
	When iterating over an ObjectList, deleted or inactive Objects are filtered out."""
	def __init__(self):
		self.changed = Signal() 
		self.added = Signal()
		self.removed = Signal()

		self.objects = []

	def __repr__(self):
		return repr(self.objects)

	def __len__(self):
		return sum(obj.active for obj in self.objects)

	def __getitem__(self, index):
		return self.objects[index]

	def __iter__(self):
		"""Iterate over all active objects in the list"""
		return (obj for obj in self.objects if obj.active)

	def index(self, index):
		return self.objects.index(index)

	def add(self, obj):
		"""Add an object to the list"""
		obj.changed.connect(self.changed)
		obj.statusChanged.connect(self.updateStatus, obj)
		self.objects.append(obj)
		if obj.active:
			self.added.emit(obj)
			self.changed.emit()

	def remove(self, obj):
		"""Remove an item from the list"""
		obj.changed.disconnect(self.changed)
		obj.statusChanged.disconnect(self.updateStatus, obj)
		self.objects.remove(obj)
		if obj.active:
			self.removed.emit(obj)
			self.changed.emit()

	def updateStatus(self, obj):
		if obj.active:
			self.added.emit(obj)
		else:
			self.removed.emit(obj)


class Node(Object):
	"""A special child class of :py:class:`~in_toolset.model.base.Object`.
	It introduces directional connections between Nodes, which allows for the creation of graphs and the like"""
	def __init__(self):
		super().__init__()
		self.preset = ObjectList()
		self.postset = ObjectList()

	def connect(self, target):
		"""Create an outgoing connection from `self` to `target`"""
		self.postset.add(target)
		target.preset.add(self)
		self.changed.emit()

	def disconnect(self, target):
		"""Remove an outgoing connection from `self` to `target` if present"""
		self.postset.remove(target)
		target.preset.remove(self)
		self.changed.emit()


class Place(Node):
	"""A representation of a place as in petri nets, implemented as a child class of :py:class:`~in_toolset.model.base.Node`.
	Can contain a positive number of uncoloured tokens,
	and can be connected to incoming or outgoing transitions transitions."""
	tokens = Property("tokensChanged", 0)

	def __init__(self):
		super().__init__()
		self.tokensChanged = Signal()
		self.tokensChanged.connect(self.changed)

	def setTokens(self, tokens):
		"""Set the number of tokens in `self` to `tokens`"""
		self.tokens = tokens

	def take(self):
		"""Decrease the number of tokens by 1"""
		self.tokens -= 1

	def give(self):
		"""Increase the number of tokens by 1"""
		self.tokens += 1


class Transition(Node):
	"""A representation of a transition as in petri nets, implemented as a child class of :py:class:`~in_toolset.model.base.Node`.
	Can be connected to an arbitrary number of places."""
	
	enabled = Property("enabledChanged", True) #:The :py:class:`~in_toolset.common.Property` `enabled` shows whether the transition is currently enabled.

	def __init__(self):
		super().__init__()
		self.enabledChanged = Signal()
		self.enabledChanged.connect(self.changed)

		self.triggered = Signal() #:The :py:class:`~in_toolset.common.Signal` `triggered` emits when the transition is triggered

		self.preset.changed.connect(self.updateEnabled)
		self.postset.changed.connect(self.updateEnabled)

	def checkEnabled(self):
		for place in self.preset:
			if place.tokens == 0:
				return False
		return True

	def updateEnabled(self):
		self.enabled = self.checkEnabled()

	def trigger(self):
		"""Trigger `self`, updating the amount of tokens contained in all connected places"""
		for place in self.preset:
			place.take()
		for place in self.postset:
			place.give()
		self.triggered.emit()


class PetriNet(Object):
	"""A representation of a petrinet, containing places and transitions, implemented as a child class of :py:class:`~in_toolset.model.base.Object`.
	Places and transitions are represented as objects of type :py:class:`~in_toolset.model.base.Place` and :py:class:`~in_toolset.model.base.Transition` , respectively"""
	
	deadlock = Property("deadlockChanged", True) #:The :py:class:`~in_toolset.common.Property` `deadlock` shows whether the petrinet is currently in deadlock. A petrinet is in deadlock if there are no enabled transitions.

	def __init__(self):
		super().__init__()

		self.deadlockChanged = Signal()
		self.triggered = Signal()

		self.places = ObjectList()
		self.transitions = ObjectList()
		self.transitions.changed.connect(self.checkDeadlock)
		self.transitions.added.connect(self.registerTransition)
		self.transitions.removed.connect(self.unregisterTransition)

	def registerTransition(self, trans):
		trans.triggered.connect(self.triggered)

	def unregisterTransition(self, trans):
		trans.triggered.disconnect(self.triggered)

	def enabledTransitions(self):
		return [t for t in self.transitions if t.enabled]

	def checkDeadlock(self):
		self.deadlock = len(self.enabledTransitions()) == 0

	def triggerRandom(self):
		"""Trigger a random transition, if possible"""
		random.choice(self.enabledTransitions()).trigger()

	def setInitialMarking(self):
		"""(Re)Set all places in `self` to the default initial marking.
			This default initial marking is defined such that all places without any incoming transitions will contain one token,
			and the others will be empty."""
		for place in self.places:
			if len(place.preset) == 0:
				place.tokens = 1
			else:
				place.tokens = 0


	def combine(self, other):
		"""Merge `other` and `self` into a new PetriNet and return it.
		This allows for bisimulation using LTSmin"""
		newSource = Place() # The new global source place.
		leftSource = Place() # The place on the path to self.
		rightSource = Place() # The place on the path to other.

		leftEnabling = Place() # The first transition on the path to self.
		rightEnabling = Place() # The first transition on the path to other.
		leftStart = Place() # The second transition on the path to self.
		rightStart = Place() # The second transition on the path to other.

		newSource.connect(leftEnabling)
		newSource.connect(rightEnabling)
		leftEnabling.connect(leftSource)
		rightEnabling.connect(leftSource)
		leftSource.connect(leftStart)
		rightSource.connect(rightStart)

		# Connect to the old source places.
		for place in self.places:
			if len(place.preset) == 0:
				leftStart.connect(place)
		for place in other.places:
			if len(place.preset) == 0:
				leftStart.connect(place)

		self.places.add(newSource)
		self.places.add(leftSource)
		self.places.add(rightSource)

		self.transitions.add(leftEnabling)
		self.transitions.add(rightEnabling)
		self.transitions.add(leftStart)
		self.transitions.add(rightStart)

		for place in other.places:
			self.places.add(place)
		for transition in other.transitions:
			self.transitions.add(transition)



import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name="in-toolset",
	version="0.1.2",
	author="Example Author",
	author_email="author@example.com",
	description="An editor and simulator for industry workflow nets (inets), a model of interorganisational workflows based on petrinets",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/jakobwuhrer/in-toolset",
	packages=setuptools.find_packages(),
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: GNU General Public License (GPL)",
		"Operating System :: OS Independent",
	],
	install_requires=[
		'pyqt5',
	],
	entry_points={
		'console_scripts': [
			'in-toolset=in_toolset.main:main',
		],
	},
	python_requires='>=3.6',
	include_package_data=True,
)




from ..model.ui import *
from .view import *
import in_toolset.config


class ArrowBase(EditorShape):
	def __init__(self, scene):
		super().__init__(scene)
		self.setZValue(-1)
		
		self.dragMode = DragMode.SPECIAL

		self.element = ShapeElement(
			"arrow", x1=0, y1=0, x2=0, y2=0, curve=0, stretch=10
		)

		self.pen = QPen()
		self.pen.setCapStyle(Qt.RoundCap)
		self.pen.setWidth(2)

		part = ShapePart()
		part.setStroke(20)
		part.setPen(self.pen)
		part.addElement(self.element)

		shape = Shape()
		shape.addPart(part)

		self.setShape(shape)
		
	def setColor(self, color):
		self.pen.setColor(color)
		
	def setCurve(self, curve):
		self.element.curve = curve
		self.updateShape()
		
	def setType(self, type):
		self.element.type = type
		self.updateShape()

	def setPoints(self, x1, y1, x2, y2):
		self.element.x1 = x1
		self.element.y1 = y1
		self.element.x2 = x2
		self.element.y2 = y2
		self.updateShape()
		
		
class ArrowItem(ArrowBase):
	def __init__(self, scene, arrow):
		super().__init__(scene)
		self.arrow = arrow
		self.connect(self.arrow.restored, self.addToScene)
		self.connect(self.arrow.deleted, self.removeFromScene)
		self.connect(self.arrow.curveChanged, self.updateArrow)
		self.connect(self.arrow.source.positionChanged, self.updateArrow)
		self.connect(self.arrow.target.positionChanged, self.updateArrow)

		self.distance = 0

		self.updateArrow()
		
	def setDistance(self, distance):
		self.distance = distance
		self.updateArrow()
		
	def delete(self):
		self.arrow.delete()
		
	def drag(self, param):
		source = self.arrow.source
		target = self.arrow.target
		
		mx = param.mouse.x()
		my = param.mouse.y()
		dx = target.x - source.x
		dy = target.y - source.y
		length = math.sqrt(dx * dx + dy * dy)
		dist = (dy * mx - dx * my + target.x * source.y - target.y * source.x) / length
		
		# This is an approximation
		angle = math.atan2(-dist * 2, length / 2)
		offset = math.sin(angle) * self.distance * 2
		if abs(offset) > abs(dist):
			offset = -dist
		curve = -dist * 2 - offset
		
		self.arrow.setCurve(curve)
		
	def updateArrow(self):
		source = self.arrow.source
		target = self.arrow.target

		dx = target.x - source.x
		dy = target.y - source.y
		length = math.sqrt(dx * dx + dy * dy)
		
		# This is an approximation
		curveAngle = math.atan2(self.arrow.curve, length / 2)
		length -= math.cos(curveAngle) * self.distance
		
		angle1 = math.atan2(dy, dx) + curveAngle
		angle2 = math.atan2(dy, dx) - curveAngle
		
		self.setCurve(self.arrow.curve)
		
		self.setPoints(
			source.x + math.cos(angle1) * self.distance,
			source.y + math.sin(angle1) * self.distance,
			target.x - math.cos(angle2) * self.distance,
			target.y - math.sin(angle2) * self.distance
		)


class LabelBase(EditorItem):
	def __init__(self, scene):
		super().__init__(scene)
		
		self.font = QFont()
		self.font.setPixelSize(16)
		self.fontMetrics = QFontMetrics(self.font)
		
		self.color = Qt.black
		self.text = ""
		
	def setFontSize(self, size):
		self.font.setPixelSize(size)
		self.updateLabel()
		
	def setColor(self, color):
		self.color = color
		self.update()
		
	def setText(self, text):
		self.text = text
		self.updateLabel()
		
	def updateLabel(self):
		self.prepareGeometryChange()
		self.update()
		
	def boundingRect(self):
		rect = self.fontMetrics.boundingRect(self.text)
		rect.moveCenter(QPoint(0, 0))
		return QRectF(rect.adjusted(-2, -2, 2, 2))
		
	def paint(self, painter, option, widget):
		pen = QPen(self.color)
		if self.isSelected():
			pen.setColor(Qt.blue)
		painter.setPen(pen)

		painter.setFont(self.font)
		painter.drawText(self.boundingRect(), Qt.AlignCenter, self.text)

		
class LabelItem(LabelBase):
	def __init__(self, scene, label):
		super().__init__(scene)

		self.dragMode = DragMode.SPECIAL
		
		self.label = label
		self.connect(self.label.restored, self.addToScene)
		self.connect(self.label.deleted, self.removeFromScene)
		self.connect(self.label.textChanged, self.updateText)
		self.connect(self.label.angleChanged, self.updatePos)
		self.connect(self.label.distanceChanged, self.updatePos)
		self.connect(self.label.positionChanged, self.updatePos)
		
		self.distMin = 0
		self.distMax = 100
		
		self.updateText()
		self.updatePos()
		
	def delete(self):
		self.label.setText("")
		
	def drag(self, param):
		dx = param.pos.x() - self.label.x
		dy = param.pos.y() - self.label.y

		dist = math.sqrt(dx * dx + dy * dy)
		dist = min(max(dist, self.distMin), self.distMax)

		self.label.setAngle(math.atan2(dy, dx))
		self.label.setDistance(dist)
		
	def setRange(self, min, max):
		self.distMin = min
		self.distMax = max
		
	def updateText(self):
		self.setText(self.label.text)
		
	def updatePos(self):
		xoffs = math.cos(self.label.angle) * self.label.distance
		yoffs = math.sin(self.label.angle) * self.label.distance
		self.setPos(self.label.x + xoffs, self.label.y + yoffs)


class LooseArrowItem(EditorShape):
	def __init__(self, scene, style, arrow):
		super().__init__(scene, style.shapes["triangle"])
		
		self.arrow = arrow
		
		self.dragMode = DragMode.SPECIAL
		
		self.label = LabelItem(scene, arrow.label)
		self.label.setColor(QColor(128, 0, 255))
		self.label.setRange(0, 60)
		scene.addItem(self.label)
		
		self.connect(self.arrow.restored, self.addToScene)
		self.connect(self.arrow.deleted, self.removeFromScene)
		self.connect(self.arrow.positionChanged, self.updateArrow)
		self.connect(self.arrow.angleChanged, self.update)
		self.connect(self.arrow.transition.typeChanged, self.update)
		
		self.updateArrow()
		
	def updateArrow(self):
		self.setPos(self.arrow.x, self.arrow.y)
		self.update()
		
	def delete(self):
		self.arrow.delete()
		
	def drag(self, param):
		if isinstance(self.arrow.node.obj, UIPetriNet):
			if self.arrow.transition.channel:
				return
		dx = param.pos.x() - self.arrow.node.x
		dy = param.pos.y() - self.arrow.node.y
		self.arrow.setAngle(math.atan2(dy, dx))
		
	def paint(self, painter, option, widget):
		angle = self.arrow.angle
		if self.arrow.transition.type == TransitionType.INPUT:
			angle += math.pi
		
		painter.save()
		painter.rotate(angle * 180 / math.pi)
		super().paint(painter, option, widget)
		painter.restore()


class NodeBase(EditorShape):
	def __init__(self, scene, shape=None):
		super().__init__(scene, shape)
		self.dragMode = DragMode.NORMAL
		
	def drag(self, param):
		self.setPos(alignToGrid(param.pos))
		
	def checkCollisions(self):
		items = self.scene.collidingItems(self)
		if any(isinstance(item, NodeBase) for item in items):
			self.setInvalid(True)
		else:
			self.setInvalid(False)
		
	def paint(self, painter, option, widget):
		super().paint(painter, option, widget)
		
		if self.invalid:
			painter.save()
			brush = QBrush(Qt.red, Qt.BDiagPattern)
			painter.setBrush(brush)
			painter.setPen(Qt.NoPen)
			painter.drawRect(self.shp.rect)
			painter.restore()
			
			
class NodeFilter:
	def __init__(self, item):
		self.base = ShapeFilter(item)
		self.flashColor = QColor(128, 128, 255)
		self.item = item
		
	def applyToPen(self, pen):
		self.base.applyToPen(pen)
		
	def applyToBrush(self, brush):
		color = mergeColors(brush.color(), self.flashColor, self.item.flashValue)
		brush.setColor(color)
		
		self.base.applyToBrush(brush)
			
			
class NodeItem(NodeBase):
	def __init__(self, scene, shape, node):
		super().__init__(scene, shape)
		
		self.filter = NodeFilter(self)
		
		self.label = LabelItem(scene, node.label)
		scene.addItem(self.label)

		self.node = node
		self.connect(self.node.restored, self.addToScene)
		self.connect(self.node.deleted, self.removeFromScene)
		self.connect(self.node.positionChanged, self.updatePos)
		
		self.flashTimer = QTimer()
		self.flashTimer.setInterval(100)
		self.flashTimer.timeout.connect(self.updateFlash)
		
		self.flashValue = 0
		
		self.updatePos()
		
	def drag(self, param):
		pos = alignToGrid(param.pos)
		self.node.move(pos.x(), pos.y())

	def delete(self):
		self.flashTimer.stop()
		self.flashValue = 0
		
		self.node.delete()

	def updatePos(self):
		self.setPos(self.node.x, self.node.y)
		
	def updateFlash(self):
		self.flashValue -= .5
		if self.flashValue <= 0:
			self.flashTimer.stop()
			self.flashValue = 0
		
		self.update()
		
	def flash(self):
		self.flashValue = 1
		self.flashTimer.start()
		self.update()



"""The :py:mod:`in_toolset.model` module contains code for modelling the behaviour of petri nets, industry nets and generic uncoloured petri nets.

It relies heavily on the signalling and property classes in :py:mod:`in_toolset.common`."""




from ..model.ui import *
from .common import *
from .scene import *
from ..common import *
import in_toolset.config
import random


class EnterpriseItem(NodeItem):
	def __init__(self, scene, style, obj):
		super().__init__(scene, style.shapes["enterprise"], obj)
		self.connect(self.node.obj.net.triggered, self.flash)


class ChannelArrowLabel(LabelBase):
	def __init__(self, scene, arrow):
		super().__init__(scene)

		self.arrow = arrow
		self.connect(self.arrow.deleted, self.removeFromScene)
		self.connect(self.arrow.curveChanged, self.updatePos)
		self.connect(self.arrow.source.node.positionChanged, self.updatePos)
		self.connect(self.arrow.target.node.positionChanged, self.updatePos)
		self.connect(self.arrow.channel.tokensChanged, self.updateText)

		self.setFontSize(18)

		self.updateText()
		self.updatePos()

	def updateText(self):
		if self.arrow.channel.tokens:
			self.setText("%i" %self.arrow.channel.tokens)
		else:
			self.setText("")

	def updatePos(self):
		source = self.arrow.source
		target = self.arrow.target

		dx = target.x - source.x
		dy = target.y - source.y
		length = math.sqrt(dx * dx + dy * dy)
		angle = math.atan2(dy, dx)

		centerx = (source.x + target.x) / 2
		centery = (source.y + target.y) / 2

		if self.arrow.curve > 0:
			distance = self.arrow.curve / 2 + 20
		else:
			distance = self.arrow.curve / 2 - 20

		posx = centerx + math.cos(angle + math.pi / 2) * distance
		posy = centery + math.sin(angle + math.pi / 2) * distance

		self.setPos(posx, posy)


class MessageAnimItem(EditorShape):
	def __init__(self, scene, style, arrow, isTarget):
		super().__init__(scene, style.shapes["messageAnim"])

		self.isTarget = isTarget

		self.arrow = arrow
		self.connect(self.arrow.deleted, self.kill)
		self.connect(self.arrow.source.positionChanged, self.updatePos)
		self.connect(self.arrow.target.positionChanged, self.updatePos)
		self.connect(self.arrow.curveChanged, self.updatePos)

		self.step = 0

		self.timer = QTimer()
		self.timer.setInterval(50)
		self.timer.timeout.connect(self.updateStep)
		self.timer.start()

		self.updatePos()

	def kill(self):
		self.timer.stop()
		self.removeFromScene()

	def updateStep(self):
		self.step += .1
		if self.step >= 1:
			self.kill()
		else:
			self.updatePos()

	def updatePos(self):
		step = self.step / 2
		if self.isTarget:
			step += .5

		source = self.arrow.source
		target = self.arrow.target

		psource = QPointF(source.x, source.y)
		ptarget = QPointF(target.x, target.y)
		center = (psource + ptarget) / 2

		diff = ptarget - psource
		length = math.sqrt(diff.x() ** 2 + diff.y() ** 2)
		angle = math.atan2(diff.y(), diff.x())

		p0 = psource
		p2 = ptarget
		p1 = QPointF(
			center.x() + math.cos(angle + math.pi / 2) * self.arrow.curve,
			center.y() + math.sin(angle + math.pi / 2) * self.arrow.curve
		)

		t1 = p0 + (p1 - p0) * step
		t2 = p1 + (p2 - p1) * step
		pos = t1 + (t2 - t1) * step

		derivative = (p1 - p0) * (1 - step) + (p2 - p1) * step
		normal = QPointF(derivative.y(), -derivative.x())
		length = math.sqrt(normal.x() ** 2 + normal.y() ** 2)

		if self.arrow.curve > 0:
			self.setPos(pos - normal / length * 20)
		else:
			self.setPos(pos + normal / length * 20)


class ChannelArrowItem(ArrowItem):
	def __init__(self, scene, style, arrow):
		super().__init__(scene, arrow)

		self.style = style

		self.setType("line")

		self.connect(self.arrow.curveChanged, self.updateAngle)
		self.connect(self.arrow.source.node.positionChanged, self.updateAngle)
		self.connect(self.arrow.target.node.positionChanged, self.updateAngle)
		self.connect(self.arrow.source.transition.triggered, self.triggerSource)
		self.connect(self.arrow.target.transition.triggered, self.triggerTarget)

	def updateAngle(self):
		if self.arrow.active:
			source = self.arrow.source
			target = self.arrow.target

			dx = target.node.x - source.node.x
			dy = target.node.y - source.node.y
			length = math.sqrt(dx * dx + dy * dy)

			angle1 = math.atan2(dy, dx)
			angle2 = math.atan2(self.arrow.curve, length / 2)
			self.arrow.source.setAngle(angle1 + angle2)
			self.arrow.target.setAngle(angle1 - angle2 + math.pi)

	def triggerSource(self):
		item = MessageAnimItem(self.scene, self.style, self.arrow, False)
		self.scene.addItem(item)

	def triggerTarget(self):
		item = MessageAnimItem(self.scene, self.style, self.arrow, True)
		self.scene.addItem(item)


class TemporaryArrowItem(ArrowBase):
	def __init__(self, scene, source):
		super().__init__(scene)

		self.setType("line")

		self.source = source

	def setTarget(self, x, y):
		if isinstance(self.source, LooseArrowItem):
			dx = x - self.source.arrow.node.x
			dy = y - self.source.arrow.node.y
			self.source.arrow.setAngle(math.atan2(dy, dx))

		sx = self.source.x()
		sy = self.source.y()
		self.setPoints(x, y, sx, sy)

	def drag(self, param):
		self.setTarget(param.mouse.x(), param.mouse.y())


class IndustryController:
	def __init__(self, style, window, industry):
		self.style = style

		self.window = window
		self.toolbar = window.toolbar
		self.scene = window.scene

		self.net = industry.net
		self.graph = industry.graph

	def startPlacement(self, pos):
		type = self.toolbar.currentTool("industry")
		if type == "enterprise":
			self.scene.setHoverEnabled(False)
			item = NodeBase(self.scene, self.style.shapes["enterprise"])
			item.setPos(alignToGrid(pos))
			return item
		elif type == "message":
			obj = self.scene.findItem(pos, EnterpriseItem, LooseArrowItem)
			if obj:
				if isinstance(obj, LooseArrowItem):
					if obj.arrow.transition.channel:
						return
				arrow = TemporaryArrowItem(self.scene, obj)
				arrow.setTarget(pos.x(), pos.y())
				return arrow

	def finishPlacement(self, pos, item):
		if isinstance(item, NodeBase):
			pos = alignToGrid(pos)
			x, y = pos.x(), pos.y()

			if not item.invalid:
				maxnum = 0
				for node in self.graph.nodes:
					prefix = node.label.text[:1]
					postfix = node.label.text[1:]
					if prefix == "e" and postfix.isdigit():
						maxnum = max(int(postfix), maxnum)

				name = "e" + str(maxnum+1)

				enterprise = UIPetriNet()
				node = UINode(enterprise)
				node.label.setText(name)
				node.move(x, y)
				self.graph.nodes.add(node)

		elif isinstance(item, TemporaryArrowItem):
			sourceItem = item.source
			targetItem = self.scene.findItem(pos, EnterpriseItem, LooseArrowItem)
			if targetItem and targetItem != sourceItem:
				source, target = self.fixTransitions(sourceItem, targetItem)

				if self.checkConnection(source, target):
					dx = target.node.x - source.node.x
					dy = target.node.y - source.node.y
					angle = math.atan2(dy, dx)

					source.setAngle(angle)
					target.setAngle(angle + math.pi)

					channel = Place()
					self.net.places.add(channel)

					arrow = UIChannelArrow(source, target, channel)
					self.graph.arrows.add(arrow)

		self.scene.setHoverEnabled(True)

	def checkConnection(self, source, target):
		if source.node == target.node:
			QMessageBox.warning(
				self.window, "Loops not allowed",
				"An enterprise may not be connected to itself."
			)
			return False

		if source.transition.channel or target.transition.channel:
			QMessageBox.warning(
				self.window, "Already connected",
				"This transition is already connected"
			)
			return False

		if not source.transition.message:
			source.transition.message = target.transition.message
		if not target.transition.message:
			target.transition.message = source.transition.message

		if source.transition.message != target.transition.message:
			QMessageBox.warning(
				self.window, "Incompatible message type",
				"Only transitions with the same message type can be connected."
			)
			return False

		return True

	def fixTransitions(self, sourceItem, targetItem):
		if isinstance(sourceItem, EnterpriseItem):
			if isinstance(targetItem, EnterpriseItem):
				source = self.addTransition(sourceItem, TransitionType.OUTPUT)
				target = self.addTransition(targetItem, TransitionType.INPUT)
				return source, target

		if isinstance(sourceItem, LooseArrowItem):
			if isinstance(targetItem, LooseArrowItem):
				if targetItem.arrow.transition.type == TransitionType.INPUT:
					return sourceItem.arrow, targetItem.arrow
				return targetItem.arrow, sourceItem.arrow

			sourceItem, targetItem = targetItem, sourceItem

		if targetItem.arrow.transition.type == TransitionType.INPUT:
			source = self.addTransition(sourceItem, TransitionType.OUTPUT)
			source.transition.message = targetItem.arrow.transition.message
			return source, targetItem.arrow

		target = self.addTransition(sourceItem, TransitionType.INPUT)
		target.transition.message = targetItem.arrow.transition.message
		return targetItem.arrow, target

	def addTransition(self, target, type):
		enterprise = target.node.obj

		transition = UITransition()
		transition.type = type
		self.net.transitions.add(transition)
		enterprise.net.transitions.add(transition)

		node = UINode(transition)
		enterprise.graph.nodes.add(node)

		arrow = UILooseArrow(node, transition)
		enterprise.graph.looseArrows.add(arrow)

		output = UILooseArrow(target.node, transition)
		self.graph.looseArrows.add(output)

		return output


class EnterpriseSettings(SettingsWidget):
	def __init__(self, obj):
		super().__init__()
		self.obj = obj
		self.connect(self.obj.positionChanged, self.updatePos)
		self.connect(self.obj.label.textChanged, self.updateLabel)

		self.setStyleSheet("font-size: 16px")

		self.x = QLabel("%i" %(obj.x / GRID_SIZE))
		self.x.setAlignment(Qt.AlignRight)
		self.y = QLabel("%i" %(obj.y / GRID_SIZE))
		self.y.setAlignment(Qt.AlignRight)

		self.label = QLineEdit(obj.label.text)
		self.label.setMaxLength(config.get("ui.max_label_size"))
		self.label.textEdited.connect(self.obj.label.setText)

		self.addField("X:", self.x)
		self.addField("Y:", self.y)
		self.addField("Name:", self.label)

	def updatePos(self):
		self.x.setText("%i" %(self.obj.x / GRID_SIZE))
		self.y.setText("%i" %(self.obj.y / GRID_SIZE))

	def updateLabel(self):
		self.label.setText(self.obj.label.text)


class LooseArrowSettings(SettingsWidget):
	def __init__(self, obj):
		super().__init__()
		self.obj = obj
		self.connect(self.obj.label.textChanged, self.updateType)
		self.connect(self.obj.transition.channelChanged, self.updateEnabled)

		self.setStyleSheet("font-size: 16px")

		self.type = QLineEdit(self.obj.label.text)
		self.type.setMaxLength(config.get("ui.max_label_size"))
		self.type.textEdited.connect(self.obj.label.setText)
		self.type.setEnabled(not self.obj.transition.channel)

		self.addField("Message type:", self.type)

	def updateType(self):
		self.type.setText(self.obj.label.text)

	def updateEnabled(self):
		self.type.setEnabled(not self.obj.transition.channel)


class IndustryScene(PetriScene):
	def __init__(self, style, window):
		super().__init__(style, window)

		self.enterpriseSelected = Signal()

	def load(self, industry):
		self.loadPetriNet(industry)

	def registerTools(self, toolbar):
		toolbar.addGroup("industry")
		toolbar.selectTool("enterprise")

	def createController(self):
		return IndustryController(self.style, self.window, self.net)

	def addNode(self, node):
		item = EnterpriseItem(self.scene, self.style, node)
		item.doubleClicked.connect(lambda: self.enterpriseSelected(node))
		self.scene.addItem(item)

	def addArrow(self, arrow):
		item = ChannelArrowItem(self.scene, self.style, arrow)
		label = ChannelArrowLabel(self.scene, arrow)
		self.scene.addItem(item)
		self.scene.addItem(label)

	def addLooseArrow(self, arrow):
		item = LooseArrowItem(self.scene, self.style, arrow)
		self.scene.addItem(item)

	def createSettingsWidget(self, items):
		classes = (EnterpriseItem, LooseArrowItem)
		filtered = [item for item in items if isinstance(item, classes)]
		if len(filtered) == 1:
			item = filtered[0]
			if isinstance(item, EnterpriseItem):
				return EnterpriseSettings(item.node)
			return LooseArrowSettings(item.arrow)
